# -*- coding: utf-8 -*-
"""
Created on Tue Jan 29 11:26:41 2013

@author: wangronin
"""

import numpy as np
# import hello as h
from .boundary_handling import boundary_handling
from scipy.stats import chi
from numpy.linalg import eigh, LinAlgError, qr, cond
from numpy.random import randn, rand, shuffle
from numpy import sqrt, eye, exp, dot, add, inf, triu, isreal, isinf,\
    ones, power, log, floor, ceil, outer, zeros, array, mod, newaxis,\
    arange, append, real, argsort, size, diag, inner, r_, linspace, atleast_2d

# My fast routines...
abs = np.abs
max = np.max
min = np.min
sum = np.sum
norm = np.linalg.norm
any = np.any
all = np.all

# TODO: modularize this function, and possiblly fit it into a optimizer class
class cma_es(object):
    """

    My toy CMA-ES... with lots of variants of mutation operators...
    TODO: complete Python __doc__ of this function
    """

    def __init__(self, dim, init_wcm, fitnessfunc, opts, sampling_method=0,
                 is_register=False, is_minimize=True, restart='IPOP'):

        self.stop_dict = {}
        self.offspring = None
        self.sel = None
        self.z = None
        self.evalcount = 0
        self.eigeneval = 0
        self.fitnessfunc = fitnessfunc if is_minimize else lambda x: -fitnessfunc(x)
        self.fitness = None
        self.fitness_rank = None
        self.sampling_method = sampling_method
        self.is_minimize = is_minimize

        # TODO: implement BIPOP
        if restart == 'IPOP':
            self.restart = restart
            self.restart_count = 0
            self.inc_popsize = 2
            self.restart_budget = opts['restart_budget'] if 'restart_budget' in opts else int(20)

        # Initialize internal strategy parameters
        self.wcm = eval(init_wcm) if isinstance(init_wcm, basestring) else init_wcm
        self.lb = eval(opts['lb']) if isinstance(opts['lb'], basestring) else opts['lb']
        self.ub = eval(opts['ub']) if isinstance(opts['ub'], basestring) else opts['ub']
        self.lb = atleast_2d(self.lb)
        self.ub = atleast_2d(self.ub)
        
        if self.lb.shape[1] != 1:
            self.lb = self.lb.T 
        if self.ub.shape[1] != 1:
            self.ub = self.ub.T 
            
        self.eval_budget = int(eval(opts['eval_budget'])) if isinstance(opts['eval_budget'],
                                                                        basestring) \
                                                          else int(opts['eval_budget'])

        self.wcm = self.wcm.reshape(-1, 1)
        self.dim = dim
        self.sigma0 = opts['sigma_init']
        self.sigma = self.sigma0
        self.f_target = opts['f_target'] if self.is_minimize else -opts['f_target']

        # Strategy parameters: Selection
        self._lambda = opts['_lambda'] if opts.has_key('_lambda') else int(4 + floor(3*log(dim)))
        if isinstance(self._lambda, basestring): self._lambda = eval(self._lambda)
        _mu_prime = (self._lambda-1) / 2.0
        self._mu = opts['_mu'] if opts.has_key('_mu') else int(ceil(_mu_prime))
        if isinstance(self._mu, basestring): self._mu = eval(self._mu)

        # TODO : new weight setting weighted recombination
        self.weights = log(_mu_prime+1.0)-log(arange(1, self._mu+1)[:, newaxis])
        self.weights = self.weights / sum(self.weights)
        self.mueff = sum(self.weights)**2. / sum(self.weights**2)

        self.wcm_old = self.wcm
        self.xopt = self.wcm
        self.fopt = np.inf

        self.pc = zeros((dim, 1))
        self.ps = zeros((dim, 1))
        self.e_vector, self.e_value = eye(dim), ones((dim, 1))
        self.C = dot(self.e_vector, self.e_value * self.e_vector.T)
        self.invsqrt_C = dot(self.e_vector, self.e_value**-1.0 * self.e_vector.T)

        # Strategy parameter: Adaptation
        # TODO: verify and update the strategy parameters
        self.cc = (4. + self.mueff/self.dim) / (self.dim+4. + 2.*self.mueff/self.dim)
        self.cs = (self.mueff+2.) / (self.dim+self.mueff+5.)
        if self._mu == 1:
            self.c_1 = min([2, self._lambda/3.])/((self.dim + 1.3)**2 + self.mueff)
            self.damps = 0.3 + 2.0*self.mueff/self._lambda + self.cs

        else:  # Original settings
            self.c_1 = 2 / ((self.dim+1.3)**2 + self.mueff)
            self.damps = 1. + 2*np.max([0, sqrt((self.mueff-1)/(self.dim+1))-1]) + self.cs
        self.c_mu = min([1 - self.c_1, 
                         2 * (self.mueff - 2. + 1./self.mueff) / ((self.dim + 2)**2 + self.mueff)])
        # TODO: verify this
    #    c_mu = 2 * (mueff-2.+1/mueff) / ((dim+2)**2+mueff)
    #    cc = 4.0 / (dim + 4.0)
    #    cs = (mueff+2.) / (dim+mueff+3.)

        # damps parameter tuning
        if opts.has_key('damps'):
            self.damps = opts['damps']
        else:
            # TODO: Parameter setting for mirrored orthogonal sampling
            # damps tuning for mirrored orthogonal sampling
            if self.sampling_method == 1:
                self.damps = 1.032 - 0.7821*self.mueff/self._lambda + self.cs

            if self.sampling_method == 4 or self.sampling_method == 8:
                if 1 < 2:
                    # damps setting for small _lambda
                    self.damps = -0.6314*(sqrt((self.mueff + .1572) / (self.dim + 1.647)) +.869) + \
                                 self.cs + 1.49
                else:
                    # damps setting for large _lambda
                    self.damps = 1.17 + 4.625 * self.mueff / self._lambda - 2.704 * self.cs
                    
            # Optimal damps setting for mirrored orthogonal sampling
            if self.sampling_method == 7:
                self.damps *= .3

        # Axuilirary variables
        self.chiN = dim**.5 * (1-1./(4*dim)+1./(21*dim**2))
        self.aux = array([])

        # rescaling constant for derandomized step-size
        self.scale = self.chiN

        # Parameters for restart heuristics and warnings
        self.is_stop_on_warning = (restart == 'IPOP' or restart == 'BIPOP')
        self.flg_warning = 0

        self.tolx = 1e-12*self.sigma
        self.tolupx = 1e3*max(self.sigma)
        self.tolfun = 1e-12
        self.nbin = int(10 + ceil(30.*self.dim/self._lambda))
        self.histfunval = zeros(self.nbin)

        # evolution history registration
        self.is_info_register = is_register
        if is_register:
            self.histsigma = zeros(self.eval_budget)
            self.hist_condition_number = zeros(self.eval_budget)
            self.hist_e_value = zeros((dim, self.eval_budget))
            self.hist_fbest = zeros(self.eval_budget)
            self.hist_xbest = zeros((self.eval_budget, self.dim))

            start = 200
            self.histindex = list(r_[0, linspace(start, self.eval_budget, 10)])
            self.histdist = zeros(size(self.histindex))
            self.ii = 0

        # if performaning pairwise selection
        self.is_pairwise_selection = (self._mu != 1 and \
            (self.sampling_method == 11 or self.sampling_method == 4 or \
            self.sampling_method == 8 or self.sampling_method == 1 or\
             self.sampling_method == 7 or self.sampling_method == 44))

    def mutation(self):
        #---------------------------- Mutation --------------------------------

        # Mirroring
        mode = self.sampling_method
        dim, _lambda, sigma, evalcount, scale, aux = self.dim, self._lambda, self.sigma, self.evalcount, self.scale, self.aux

        if mode == 1 or mode == 11:
            if mod(evalcount+_lambda, 2) != 0:
                half = int(ceil(_lambda/2.))
                z = randn(dim, half)
                aux = -z[:, -1].reshape(-1, 1)
                z = append(z, -z[:, :-1], axis=1)
            else:
                half = int(floor(_lambda/2.))
                z = randn(dim, half)
                z = append(z, -z, axis=1)
                if len(aux) != 0:
                    z = append(aux, z, axis=1)
            self.half = half
        # Derandomized step size
        elif mode == 3:
            z = randn(dim, _lambda)
            z = scale * z / sqrt(sum(power(z, 2), 0))

        # Orthogonal mirrored sampling
        elif mode == 4 or mode == 7 or mode == 44:
            if mod(evalcount+_lambda, 2) != 0:
                half = int(ceil(_lambda/2.))
                z = zeros((dim, half))
                n = int(min([dim, half]))
                if dim < half:
                    z[:, dim:] = randn(dim, half-dim)
                q = qr(randn(dim, n))[0]
                l = chi.rvs(dim, size=n)
                z[:, 0:n] = l * q
                aux = -z[:, -1].reshape(-1, 1)
                z = append(z, -z[:, :-1], axis=1)
            else:
                half = int(floor(_lambda/2.))
                z = zeros((dim, half))
                n = int(min([dim, half]))
                if dim < half:
                    z[:, dim:] = randn(dim, half-dim)
                q = qr(randn(dim, n))[0]
                l = chi.rvs(dim, size=n)
                z[:, 0:n] = l * q
                z = append(z, -z, axis=1)
                if len(aux) != 0:
                    z = append(aux, z, axis=1)
            self.half = half
        # Orthogonal mirrored sampling...well
        elif mode == 8:
            if mod(evalcount+_lambda, 2) != 0:
                half = ceil(_lambda/2.)
                z = zeros((dim, half))
                n = min([dim, half])
                if dim < half:
                    z[:, dim:] = randn(dim, half-dim)
                tmp = randn(dim, n)
                l = sqrt(np.sum(power(tmp, 2), 0))
                q = qr(tmp)[0]
                z[:, 0:n] = l * q
                aux = -z[:, -1].reshape(-1, 1)
                z = append(z, -z[:, :-1], axis=1)
            else:
                half = floor(_lambda/2.)
                z = zeros((dim, half))
                n = min([dim, half])
                if dim < half:
                    z[:, dim:] = randn(dim, half-dim)
                tmp = randn(dim, n)
                l = sqrt(np.sum(power(tmp, 2), 0))
                q = qr(tmp)[0]
                z[:, 0:n] = l * q
                z = append(z, -z, axis=1)
                if len(aux) != 0:
                    z = append(aux, z, axis=1)
            self.half = half
        # Orthogonal sampling (random rotation)
        elif mode == 5:
            pass
            # z = dot(rand_orth_mat(dim), eye(dim))
            # n = dim
            # if dim > _lambda:
            #     p = arange(0, dim)
            #     shuffle(p)
            #     z = z[:, p[0:_lambda]]
            #     n = _lambda
            # l = chi.rvs(dim, size=n)
            # sign = rand(n)
            # sign[sign > .5] = 1
            # sign[sign <= .5] = -1
            # z = sign * l * z
            # if dim < _lambda:
            #     ss = randn(dim, _lambda-dim)
            #     z = append(z, ss, axis=1)

        # Orthogonal sampling (Gram-Schmidt)
        elif mode == 6:
            z = zeros((dim, _lambda))
            n = min([dim, _lambda])
            if dim < _lambda:
                z[:, dim:] = randn(dim, _lambda-dim)
            q = qr(randn(dim, n))[0]
            l = chi.rvs(dim, size=n)

        elif mode == 9:
            if mod(evalcount+_lambda, 2) != 0:
                half = int(ceil(_lambda/2.))
                z = randn(dim, half)
                q = randn(dim, half)
                p = array([q[:, i] - inner(q[:, i], z[:, i])*z[:, i] / norm(z[:, i])**2 \
                    for i in range(half)]).T
                p = (p / sqrt(sum(power(p, 2), 0))) * sqrt(sum(power(q, 2), 0))
                aux = p[:, -1].reshape(-1, 1)
                z = append(z, p[:, :-1], axis=1)
            else:
                half = int(floor(_lambda/2.))
                z = randn(dim, half)
                q = randn(dim, half)
                p = array([q[:, i] - inner(q[:, i], z[:, i])*z[:, i] / norm(z[:, i])**2 \
                    for i in range(half)]).T
                p = (p / sqrt(sum(power(p, 2), 0))) * sqrt(sum(power(q, 2), 0))
                z = append(z, p, axis=1)
                if len(aux) != 0:
                    z = append(aux, z, axis=1)
            self.half = half
        # Standard mutation
        else:
            z = randn(dim, _lambda)

        self.z = z
        self.offspring = add(self.wcm, sigma * self.e_vector.dot(self.e_value * self.z))

    def constraint_handling(self):
        self.offspring = boundary_handling(self.offspring, self.lb, self.ub)

    def evaluation(self):
        try:
            self.fitness = self.fitnessfunc(self.offspring)
        except Exception:  # if the fitness function evaluates a single point each time
            self.fitness = np.array([self.fitnessfunc(_) for _ in self.offspring.T])
        self.evalcount += self._lambda
        self.fitness_rank = argsort(self.fitness)
        self.fitness_true = self.fitness * (-1) ** (~self.is_minimize)

    def update(self):
        # Cumulation: Update evolution paths
        cc, cs, c_1, c_mu = self.cc, self.cs, self.c_1, self.c_mu
        wcm, wcm_old, mueff, invsqrt_C = self.wcm, self.wcm_old, self.mueff, self.invsqrt_C
        evalcount, _lambda = self.evalcount, self._lambda

        self.ps = (1-cs) * self.ps + sqrt(cs*(2-cs)*mueff) \
            * dot(invsqrt_C, (wcm - wcm_old) / self.sigma)
        hsig = sum(self.ps**2.)/(1.-(1.-cs)**(2.*evalcount/_lambda))/self.dim < 2. + 4./(self.dim+1.)
        self.pc = (1-cc) * self.pc + \
            hsig * sqrt(cc*(2.-cc)*mueff) * (wcm - wcm_old) / self.sigma

        offset = (self.offspring[:, self.sel] - wcm_old) / self.sigma
        self.C = (1.0-c_1-c_mu) * self.C \
            + c_1 * (outer(self.pc, self.pc) \
                + (1.-hsig) * cc*(2-cc) * self.C) \
            + c_mu * dot(offset, self.weights*offset.T)
        # Adapt step size sigma
        self.sigma = self.sigma * exp((norm(self.ps)/self.chiN - 1) * self.cs/self.damps)
        
        if 11 < 3: # TODO: Saw in the Hansen's codde, reason for this...
            self.sigma = self.sigma * exp(min([1, (norm(self.ps)/self.chiN - 1) * \
                self.cs/self.damps]))

    def updateBD(self):
        # Eigen decomposition
        C = self.C # lastest setting for
        C = triu(C) + triu(C, 1).T                  # eigen decomposition
        if any(isinf(C)) > 1:                           # interval
            self.flg_warning ^= 2**0
        else:
            try:
                w, e_vector = eigh(C)
                e_value = sqrt(map(complex, w)).reshape(-1, 1)
                if any(~isreal(e_value)) or any(isinf(e_value)):
                    if self.is_stop_on_warning:
                        self.stop_dict['EigenvalueError'] = True
                    else:
                        self.flg_warning ^= 2**1
                else:
                    self.e_value = real(e_value)
                    self.e_vector = e_vector
                    self.invsqrt_C = dot(e_vector, e_value**-1 * e_vector.T)
            except LinAlgError:
                if self.is_stop_on_warning:
                    self.stop_dict['linalgerror'] = True
                else:
                    self.flg_warning ^= 2**1

    def info_register(self):
        # Register historical info
        evalcount, _lambda, fitness, offspring = self.evalcount, \
            self._lambda, self.fitness, self.offspring
        if self.is_info_register:
            if self.ii < len(self.histindex):
                if abs(self.evalcount - self.histindex[self.ii]) <= self._lambda:
                    self.histdist[self.ii] = norm(self.wcm_old)
                    self.ii += 1
            self.histsigma[evalcount-_lambda:evalcount] = self.sigma
            self.hist_condition_number[evalcount-_lambda:evalcount] = max(self.e_value)**2 / \
                min(self.e_value) ** 2.
            self.hist_e_value[:, evalcount-_lambda:evalcount] = self.e_value
            self.hist_xbest[evalcount-_lambda:evalcount, :] = offspring[:, self.sel[0]]
            self.hist_fbest[evalcount-_lambda:evalcount] = fitness[self.sel[0]]

    def check_stop_criteria(self):
        #-------------------------- Restart criterion ------------------------------
        is_stop_on_warning = self.is_stop_on_warning
        sigma, evalcount, _lambda, fitness = self.sigma, \
            self.evalcount, self._lambda, self.fitness

        self.stop_dict['ftarget'] = True if self.fopt <= self.f_target else False
        self.stop_dict['maxfevals'] = True if self.evalcount >= self.eval_budget else False

        if self.evalcount != 0:
            if np.any(fitness == inf) or np.any(fitness == np.nan):
                # TODO: nasty error to be debugged
                raise Exception('Somthing is wrong!')

            if (sigma < 1e-16) or (sigma > 1e6):
                self.flg_warning = True

            diagC = diag(self.C).reshape(-1, 1)
            self.histfunval[mod(evalcount/_lambda-1, self.nbin)] = fitness[self.sel[0]]
            if mod(evalcount, _lambda) == self.nbin and \
                max(self.histfunval) - min(self.histfunval) < self.tolfun:
                if is_stop_on_warning:
                    self.stop_dict['tolfun'] = True
                else:
                    self.flg_warning = True

            # Condition covariance
            if cond(self.C) > 1e14:
                if is_stop_on_warning:
                    self.stop_dict['conditioncov'] = True
                else:
                    self.flg_warning = True

            # TolX
            tmp = append(abs(self.pc), sqrt(diagC), axis=1)
            if all(self.sigma*(max(tmp, axis=1)) < self.tolx):
                if is_stop_on_warning:
                    self.stop_dict['TolX'] = True
                else:
                    self.flg_warning = True

            # TolUPX
            if any(sigma*sqrt(diagC)) > self.tolupx:
                if is_stop_on_warning:
                    self.stop_dict['TolUPX'] = True
                else:
                    self.flg_warning = True

            # No effective axis
            a = mod(evalcount/_lambda-1, self.dim)
            if all(0.1*sigma*self.e_value[a, 0]*self.e_vector[:, a] + self.wcm == self.wcm):
                if is_stop_on_warning:
                     self.stop_dict['noeffectaxis'] = True
                else:
                    sigma *= exp(0.2 + self.cs / self.damps)

            # No effective coordinate
            if any(0.2*sigma*sqrt(diagC) + self.wcm == self.wcm):
                if is_stop_on_warning:
                    self.stop_dict['noeffectcoord'] = True
                else:
                    self.C += (self.c_1 + self.c_mu) * diag(diagC * \
                        (self.wcm == self.wcm + 0.2*sigma*sqrt(diagC)))
                    sigma *= exp(0.05 + self.cs/self.damps)
            # Adjust step size in case of equal function values
            if fitness[self.sel[0]] == fitness[self.sel[int(min([ceil(0.1+_lambda/4.0), 
                                               self._mu-1]))]]:
                if is_stop_on_warning:
                    self.stop_dict['flatfitness'] = True
                else:
                    sigma *= exp(0.2+self.cs/self.damps)

        # Handling warnings: Internally rectification of strategy paramters
        if self.flg_warning != 0:
            self.reset_state()
            self.flg_warning = False

    def reset_state(self):
        self.C = eye(self.dim)
        self.e_vector = eye(self.dim)
        self.e_value = ones((self.dim, 1))
        self.invsqrt_C = eye(self.dim)
        self.pc = zeros((self.dim, 1))
        self.ps = zeros((self.dim, 1))
        self.sigma = self.sigma0

    def reset_stop_dict(self):
        for key, _ in self.stop_dict.iteritems():
            self.stop_dict[key] = False

    def optimize(self):
        # TODO: use IPOP for now, to implement BIPOP method
        while self.restart_count < self.restart_budget:
            while True:
                self.info_register()
                self.mutation()
                self.constraint_handling()
                self.evaluation()

                #--------------------------- Comma selection ----------------------------
                # pairwise selection for mirroring
                if self.is_pairwise_selection:
#                    self.sel = pairwise_selection(self.fitness_rank, self.half, self._mu)
                    pass
                else:
                    self.sel = self.fitness_rank[0:self._mu]

                # ------------------------- Weighted recombination ----------------------
                self.wcm_old = self.wcm
                self.wcm = dot(self.offspring[:, self.sel], self.weights)
                self.update()

                # update the eigenvectors and eigenvalues. for computational time concern
                if self.evalcount - self.eigeneval > self._lambda/(self.c_1+self.c_mu)/self.dim/10:
                    self.eigeneval = self.evalcount
                    self.updateBD()

                if self.fopt > self.fitness[self.sel[0]]:
                    self.fopt = self.fitness[self.sel[0]]
                    self.xopt = self.offspring[:, self.sel[0]].reshape(self.dim, -1)

                self.check_stop_criteria()

                if any(array(self.stop_dict.values()) == True):
                    break

            # if the termination criteria are met...
            if self.stop_dict['ftarget'] or self.stop_dict['maxfevals']:
                break
            
            self.restart_count += 1
            self._lambda *= self.inc_popsize
            _mu_prime = (self._lambda-1) / 2.0
            self._mu = int(ceil(_mu_prime))
            self.weights = log(_mu_prime+1.0)-log(arange(1, self._mu+1)[:, newaxis])
            self.weights = self.weights / sum(self.weights)
            self.mueff = sum(self.weights)**2. / sum(self.weights**2)

            self.wcm = rand(self.dim, 1) * (self.ub - self.lb) + self.lb
            self.reset_state()
            self.reset_stop_dict()

        return self.xopt, self.fopt, self.evalcount, self.stop_dict


# -*- coding: utf-8 -*-
"""
Created on Thu Sep  7 11:10:18 2017

@author: wangronin
"""
import numpy as np
from numpy import exp, nonzero, argsort, ceil, zeros, mod
from numpy.random import randint, rand, randn, geometric

from ..misc import handle_box_constraint
from ..base import Solution
from ..utils import dynamic_penalty
from ..SearchSpace import ContinuousSpace, OrdinalSpace, NominalSpace

# TODO: test more contraint handling methods
# TODO: improve efficiency, e.g. compile it with cython
# TODO: try to use advanced python default parameter 
class MIES(object):
    def __init__(self, search_space, obj_func, eq_func=None, ineq_func=None, x0=None, 
                 ftarget=None, max_eval=np.inf, minimize=True, elitism=False, mu_=4, 
                 lambda_=10, sigma0=None, eta0=None, P0=None, verbose=False,
                 eval_type='list'):

        self.mu_ = mu_
        self.lambda_ = lambda_
        self.eval_count = 0
        self.iter_count = 0
        self.minimize = minimize
        self.obj_func = obj_func
        self.eq_func = eq_func
        self.ineq_func = ineq_func
        self.stop_dict = {}
        self.verbose = verbose
        self.max_eval = max_eval
        self.ftarget = ftarget
        self.elitism = False
        self._penalty_func = dynamic_penalty
        self._eval_type = eval_type
        
        self._space = search_space
        self.var_names = self._space.var_name
        self.param_type = self._space.var_type

        # index of each type of variables in the dataframe
        self.id_r = self._space.id_C       # index of continuous variable
        self.id_i = self._space.id_O       # index of integer variable
        self.id_d = self._space.id_N       # index of categorical variable

        # the number of variables per each type
        self.N_r = len(self.id_r)
        self.N_i = len(self.id_i)
        self.N_d = len(self.id_d)
        self.dim = self.N_r + self.N_i + self.N_d

        # by default, we use individual step sizes for continuous and 
        # integer variables and global strength for the nominal variables
        self.N_p = min(self.N_d, int(1))
        
        # total length of the solution vector
        self._len = self.dim + self.N_r + self.N_i + self.N_p 
        
        # unpack interval bounds
        self.bounds_r = np.asarray([self._space.bounds[_] for _ in self.id_r])
        self.bounds_i = np.asarray([self._space.bounds[_] for _ in self.id_i])
        self.bounds_d = np.asarray([self._space.bounds[_] for _ in self.id_d])
        self._check_bounds(self.bounds_r)
        self._check_bounds(self.bounds_i)
        
        # step default step-sizes/mutation strength
        par_name = []
        if sigma0 is None and self.N_r:
            sigma0 = 0.05 * (self.bounds_r[:, 1] - self.bounds_r[:, 0])
            par_name += ['sigma' + str(_) for _ in range(self.N_r)]
        if eta0 is None and self.N_i:
            eta0 = 0.05 * (self.bounds_i[:, 1] - self.bounds_i[:, 0]) 
            par_name += ['eta' + str(_) for _ in range(self.N_i)]
        if P0 is None and self.N_d:
            P0 = 1. / self.N_d
            par_name += ['P' + str(_) for _ in range(self.N_p)]

        # column indices: used for slicing
        self._id_var = np.arange(self.dim)                    
        self._id_sigma = np.arange(self.N_r) + len(self._id_var)
        self._id_eta = np.arange(self.N_i) + len(self._id_var) + len(self._id_sigma) 
        self._id_p = np.arange(self.N_p) + len(self._id_var) + len(self._id_sigma) \
            + len(self._id_eta)
        self._id_hyperpar = np.arange(self.dim, self._len)

         # initialize the populations
        if x0 is not None:                         # given x0
            par = []
            if self.N_r:
                par += [sigma0]
            if self.N_i:
                par += [eta0]
            if self.N_p:
                par += [P0] * self.N_p

            self.pop = Solution(np.tile(np.r_[x0, par], (self.mu_, 1)),
                                var_name=self.var_names + par_name, 
                                verbose=self.verbose)
            fitness0 = self.evaluate(self.pop[0])
            self.fitness = np.repeat(fitness0, self.mu_)
            self.xopt = x0
            self.fopt = sum(fitness0)
        else:                                      # uniform sampling                 
            x = np.asarray(self._space.sampling(self.mu_), dtype='object')   

            par = []
            if self.N_r:
                par += [np.tile(sigma0, (self.mu_, 1))]
            if self.N_i:
                par += [np.tile(eta0, (self.mu_, 1))]
            if self.N_p:
                par += [np.tile([P0] * self.N_p, (self.mu_, 1))]

            par = np.concatenate(par, axis=1)
            x = np.c_[x, par].tolist()
            
            self.pop = Solution(x, var_name=self.var_names + par_name, verbose=self.verbose)
            self.fitness = self.evaluate(self.pop)
            self.fopt = min(self.fitness) if self.minimize else max(self.fitness)
            _ = np.nonzero(self.fopt == self.fitness)[0][0]
            self.xopt = self.pop[_, self._id_var]
        
        self.offspring = self.pop[0] * self.lambda_
        self.f_offspring = np.repeat(self.fitness[0], self.lambda_)
        self._set_hyperparameter()

        # stopping criteria
        self.tolfun = 1e-5
        self.nbin = int(3 + ceil(30. * self.dim / self.lambda_))
        self.histfunval = zeros(self.nbin)
    
    def _check_bounds(self, bounds):
        if len(bounds) == 0:
            return
        if any(bounds[:, 0] >= bounds[:, 1]):
            raise ValueError('lower bounds must be smaller than upper bounds')

    def _set_hyperparameter(self):
        # hyperparameters: mutation strength adaptation
        if self.N_r:
            self.tau_r = 1 / np.sqrt(2 * self.N_r)
            self.tau_p_r = 1 / np.sqrt(2 * np.sqrt(self.N_r))

        if self.N_i:
            self.tau_i = 1 / np.sqrt(2 * self.N_i)
            self.tau_p_i = 1 / np.sqrt(2 * np.sqrt(self.N_i))

        if self.N_d:
            self.tau_d = 1 / np.sqrt(2 * self.N_d)
            self.tau_p_d = 1 / np.sqrt(2 * np.sqrt(self.N_d))

    def recombine(self, id1, id2):
        p1 = self.pop[id1].copy()  # IMPORTANT: make a copy
        if id1 != id2:
            p2 = self.pop[id2]
            # intermediate recombination for the mutation strengths
            p1[self._id_hyperpar] = (np.array(p1[self._id_hyperpar]) + \
                np.array(p2[self._id_hyperpar])) / 2

            # dominant recombination for solution parameters
            _, = np.nonzero(randn(self.dim) > 0.5)
            p1[_] = p2[_]
        return p1

    def select(self):
        pop = self.pop + self.offspring if self.elitism else self.offspring
        fitness = np.r_[self.fitness, self.f_offspring] if self.elitism else self.f_offspring
        rank = argsort(fitness)

        if not self.minimize:
            rank = rank[::-1]
        
        _ = rank[:self.mu_]
        self.pop = pop[_]
        self.fitness = fitness[_]

    def evaluate(self, pop, return_penalized=True):
        if len(pop.shape) == 1:  # one solution
            pop.fitness = np.asarray(self.obj_func(pop[self._id_var]))
        else:                    # a population
            pop.fitness = np.array(list(map(self.obj_func, pop[:, self._id_var])))
        
        self.eval_count += pop.N
        if self._eval_type == 'list':
            X = pop
        elif self._eval_type == 'dict':
            X = [self._space.to_dict(x) for x in pop]

        _penalized_fitness = pop.fitness + \
            self._penalty_func(
                X, self.iter_count + 1, 
                self.eq_func, self.ineq_func, 
                minimize=self.minimize
            )
        return (_penalized_fitness if return_penalized else pop.fitness)

    def mutate(self, individual):
        if self.N_r:
            self._mutate_r(individual)
        if self.N_i:
            self._mutate_i(individual)
        if self.N_d:
            self._mutate_d(individual)
        return individual

    def _mutate_r(self, individual):
        sigma = np.asarray(individual[self._id_sigma], dtype='float')
        # mutate step-sizes
        if len(self._id_sigma) == 1:
            sigma = sigma * exp(self.tau_r * randn())
        else:
            sigma = sigma * exp(self.tau_r * randn() + self.tau_p_r * randn(self.N_r))
        
        # Gaussian mutation
        R = randn(self.N_r)
        x = np.asarray(individual[self.id_r], dtype='float')
        x_ = x + sigma * R
        
        # Interval Bounds Treatment
        x_ = handle_box_constraint(x_, self.bounds_r[:, 0], self.bounds_r[:, 1])
        
        # TODO: check if this interval boundary handling works with penalty functions
        # the constraint handling method will (by chance) turn really bad cadidates
        # (the one with huge sigmas) to good ones and hence making the step size explode
        # Repair the step-size if x_ is out of bounds
        if 11 < 2:
            individual[self._id_sigma] = np.abs((x_ - x) / R)
        else:
            individual[self._id_sigma] = sigma
        individual[self.id_r] = x_
        
    def _mutate_i(self, individual):
        eta = np.asarray(individual[self._id_eta].tolist(), dtype='float')
        x = np.asarray(individual[self.id_i], dtype='int')
        if len(self._id_eta) == 1:
            eta = eta * exp(self.tau_i * randn())
        else:
            eta = eta * exp(self.tau_i * randn() + self.tau_p_i * randn(self.N_i))
        eta[eta > 1] = 1

        p = 1 - (eta / self.N_i) / (1 + np.sqrt(1 + (eta / self.N_i) ** 2.))
        x_ = x + geometric(p) - geometric(p)

        # Interval Bounds Treatment
        x_ = np.asarray(handle_box_constraint(x_, self.bounds_i[:, 0], self.bounds_i[:, 1]), dtype='int')

        individual[self.id_i] = x_
        individual[self._id_eta] = eta

    def _mutate_d(self, individual):
        P = np.asarray(individual[self._id_p], dtype='float')
        #  Unbiased mutation on the mutation probability
        P = 1. / (1. + (1. - P) / P * exp(-self.tau_d * randn()))
        individual[self._id_p] = handle_box_constraint(P, 1. / (3. * self.N_d), 0.5)

        idx, = np.nonzero(rand(self.N_d) < P)
        for i in idx:
            levels = self.bounds_d[i]
            individual[self.id_d[i]] = levels[randint(0, len(levels))]

    def stop(self):
        if self.eval_count > self.max_eval:
            self.stop_dict['max_eval'] = True

        if self.eval_count != 0 and self.iter_count != 0:
            fitness = self.f_offspring
            # sigma = np.atleast_2d([__[self._id_sigma] for __ in self.pop]) 
            # sigma_mean = np.mean(sigma, axis=0)
            
            # tolerance on fitness in history
            self.histfunval[int(mod(self.eval_count / self.lambda_ - 1, self.nbin))] = fitness[0]
            if mod(self.eval_count / self.lambda_, self.nbin) == 0 and \
                (max(self.histfunval) - min(self.histfunval)) < self.tolfun:
                    self.stop_dict['tolfun'] = True
            
            # flat fitness within the population
            if fitness[0] == fitness[int(min(ceil(.1 + self.lambda_ / 4.), self.mu_ - 1))]:
                self.stop_dict['flatfitness'] = True
            
            # TODO: implement more stop criteria
            # if any(sigma_mean < 1e-10) or any(sigma_mean > 1e10):
            #     self.stop_dict['sigma'] = True

            # if cond(self.C) > 1e14:
            #     if is_stop_on_warning:
            #         self.stop_dict['conditioncov'] = True
            #     else:
            #         self.flg_warning = True

            # # TolUPX
            # if any(sigma*sqrt(diagC)) > self.tolupx:
            #     if is_stop_on_warning:
            #         self.stop_dict['TolUPX'] = True
            #     else:
            #         self.flg_warning = True
            
        return any(self.stop_dict.values())

    def _better(self, f1, f2):
        return f1 < f2 if self.minimize else f1 > f2

    def optimize(self):
        while not self.stop():
            # TODO: vectorize this part
            for i in range(self.lambda_):
                p1, p2 = randint(0, self.mu_), randint(0, self.mu_)
                individual = self.recombine(p1, p2)
                self.offspring[i] = self.mutate(individual)
            
            # NOTE: `self.fitness` and `self.f_offspring` are penalized function values
            self.f_offspring[:] = self.evaluate(self.offspring)
            self.select()

            curr_best = self.pop[0]
            xopt_, fopt_ = curr_best[self._id_var], self.fitness[0]

            if self._better(fopt_, self.fopt):
                self.xopt, self.fopt = xopt_, fopt_
            self.iter_count += 1

            if self.verbose:
                print('iteration {}, fopt: {}'.format(self.iter_count + 1, self.fopt))
                print(self.xopt)

        self.stop_dict['funcalls'] = self.eval_count
        return self.xopt.tolist(), self.fopt, self.stop_dict


"""
Created on Mon Apr 23 17:16:39 2018

@author: Hao Wang
@email: wangronin@gmail.com

"""
import numpy as np
from tabulate import tabulate
from typing import Callable, Any, Tuple

# TODO: test the performance against Pandas dataframe
# TODO: maybe set var_name as a constant attribute
# TODO: register the objective function the class and implement an eval function  
# TODO: fix the maximum recursion problem when debugging

class Solution(np.ndarray):
    """Subclassing numpy array to represent set of solutions in the optimization
       Goal to achieve: 
        1) heterogenous data types, like pandas
        2) easy indexing as np.ndarray
        3) extra attributes (e.g., fitness) sliced together with the solution
    """
    # TODO: get rid of self.__dict__ for speed concern
    # But __slots__ does NOT work with dill yet...
    # __slots__ = 'N', 'dim', 'var_name', 'index', 'fitness', 'n_eval', 'verbose', 'n_obj'
    def __new__(cls, x, fitness=None, n_eval=0, index=None, var_name=None, 
                fitness_name=None, n_obj=1, verbose=True):
        """
        Parameters
        ----------
            x : array-like,
            fitness : float (array),
                objective values of solutions
            n_eval : int (array),
                number of evaluations per each solution           
            index : int (array),
                indices of solutions
            var_name : str (array),
                column names of variables
            fitness_name : str (array),
                column names of fitness values
            verbose : bool,
                controls if additional information are printed when calling __str__
                and to_dict
        Note
        ----
            Instead of using `__init__`, the `__new__` function is used here because 
            sometimes we would like to return an object of its subclasses, e.g., when slicing a subclass of `ndarray`, `ndarray.__new__(subclass, ...)` will return an object of type `subclass` while `ndarray.__init__(self, ...)` will return an object of `ndarray` (of course, `__init__` would work if the user also overloads the slicing function, which is not convenient).
            If attributes `index`, `fitness`, `n_eval` are modified in a slice of Solution, 
            the corresponding attributes in the original object are also modified. 
            `var_name` is not affected by this behavior. 
            This function is only called when explicitly constructing the `Solution` object. For slicing and view casting, the extra attributes are handled in function
            `__array_finalize__`.
        """
        obj = np.asarray(x, dtype='object').view(cls)

        if len(obj.shape) > 2:
            raise Exception('More than 2D is not supported')
        
        obj.N = 1 if len(obj.shape) == 1 else obj.shape[0]
        obj.dim = obj.shape[0] if len(obj.shape) == 1 else obj.shape[1]
        obj.n_obj = int(n_obj)
        
        if obj.n_obj > 1:  # multi-objective 
            if not hasattr(fitness, '__iter__'):
                fitness = [[fitness] * obj.n_obj] * obj.N
            elif not hasattr(fitness[0], '__iter__'):
                assert len(fitness) == obj.n_obj
                fitness = [fitness] * obj.N
            assert all([len(_) == obj.n_obj for _ in fitness])
        elif obj.n_obj == 1:
            if not hasattr(fitness, '__iter__'):
                fitness = [fitness] * obj.N
        assert len(fitness) == obj.N

        if not hasattr(n_eval, '__iter__'):
            n_eval = [n_eval] * obj.N
            
        if index is None:
            index = list(range(obj.N))
        elif isinstance(index, int):
            index = [index]
        assert len(index) == obj.N
        
        if var_name is None:
            if obj.dim == 1:
                var_name = ['x']
            else:
                var_name = ['x' + str(i) for i in range(obj.dim)]
        assert len(var_name) == obj.dim

        if fitness_name is None:
            if obj.n_obj == 1:
                fitness_name = ['f']
            else:
                fitness_name = ['f' + str(i) for i in range(obj.n_obj)]
        elif isinstance(fitness_name, str):
            assert obj.n_obj == 1
            fitness_name = [fitness_name]
        assert len(fitness_name) == obj.n_obj

        # a np.ndarray is used for those attributes because slicing it returns references
        # avoid calling self.__setattr__ for attributes fitness, n_eval and index
        super(Solution, obj).__setattr__('fitness', np.asarray(fitness, dtype='float'))
        super(Solution, obj).__setattr__('n_eval', np.asarray(n_eval, dtype='int'))
        super(Solution, obj).__setattr__('index', np.asarray(index, dtype='int'))
        obj.var_name = np.asarray(var_name)
        obj.fitness_name = fitness_name
        obj.verbose = verbose
        
        return obj      

    def _check_attr(self):
        # TODO: check for the compatibility of fitness and n_eval
        pass

    def __iadd__(self, other):
        return self.__add__(other)

    def __add__(self, other):
        """
        Concatenate two sets of solutions 
        """
        assert isinstance(other, Solution)
        assert self.dim == other.dim
        assert self.n_obj == other.n_obj
        assert len(set(self.fitness_name).symmetric_difference(other.fitness_name)) == 0
        assert len(set(self.var_name).symmetric_difference(other.var_name)) == 0

        _ = [self.tolist()] if len(self.shape) == 1 else self.tolist()
        __ = [other.tolist()] if len(other.shape) == 1 else other.tolist()
        return Solution(_ + __,  self.fitness.tolist() + other.fitness.tolist(),
                        self.n_eval.tolist() + other.n_eval.tolist(), 
                        var_name=self.var_name, fitness_name=self.fitness_name,
                        n_obj=self.n_obj, 
                        verbose=self.verbose)

    def __mul__(self, N):
        """
        repeat a solution N times
        """
        assert isinstance(N, int)
        if self.N > 1:
            raise Exception('Replication is not supported for 2D')
        return Solution([self.tolist()] * N, self.fitness.tolist() * N, 
                         self.n_eval.tolist() * N, var_name=self.var_name, 
                         fitness_name=self.fitness_name,
                         n_obj=self.n_obj, verbose=self.verbose)

    def __rmul__(self, N):
        return self.__mul__(N)
    
    def __setattr__(self, name, value):
        attr = getattr(self, name, None)
        if hasattr(attr, '__iter__') and name in ['fitness', 'n_eval', 'index']:
            attr[:] = value  # IMPORTANT: copy the value (not reference) to the attribute
        else:
            super(Solution, self).__setattr__(name, value)

    def __setitem__(self, index, value):
        # TODO: maybe add variable type checker here...
        super(Solution, self).__setitem__(index, value)
    
    def __getitem__(self, index):
        _, __ = index, slice(None, None)
        if isinstance(index, tuple):
            _ = index[0]
            if len(index) == 2:
                if isinstance(index[1], int) and not isinstance(index[0], int):
                    __ = slice(index[1], index[1] + 1)
                    index = (_, __)
                else:
                    __ = index[1]

        subarr = super(Solution, self).__getitem__(index)

        # sub-slicing the attributes
        if isinstance(subarr, Solution):
            # NOTE: `slice` is needed here to make sure an array is always returned 
            # after slicing attributes `fitness`, `n_eval, `index`
            # Otherwise setting the attribute of a slice is by value...
            _ = slice(_, _ + 1) if isinstance(_, (int, np.int_)) else _
            if len(self.shape) == 1:   # `self` is a 1-d array
                subarr.var_name = subarr.var_name[_]
            else:
                # NOTE: 1-d array should have 1-d `fitness`
                fitness = subarr.fitness[_]
                if len(subarr.shape) == 1:
                    fitness = fitness.ravel()
                    
                super(Solution, subarr).__setattr__('fitness', fitness)
                super(Solution, subarr).__setattr__('n_eval', subarr.n_eval[_])
                super(Solution, subarr).__setattr__('index', subarr.index[_])

                # multiple solutions and the column is indexed
                subarr.var_name = subarr.var_name[__]

            subarr.N = 1 if len(subarr.shape) == 1 else subarr.shape[0]
            subarr.dim = subarr.shape[0] if len(subarr.shape) == 1 else subarr.shape[1]
        
        return subarr

    def unique(self):
        _, index = np.unique(self.tolist(), axis=0, return_index=True)
        return self[np.sort(index)]

    def __array_finalize__(self, obj):
        """
        `__array_finalize__` is called after new `Solution` instance is created: from calling
        1) `__new__`, 2) view casting (`ndarray`.`view()`) or 3) slicing (`__getitem__`) 
        """
        if obj is None: return
        # Needed for array slicing (__getitem__)
        super(Solution, self).__setattr__('fitness', getattr(obj, 'fitness', None))
        super(Solution, self).__setattr__('n_eval', getattr(obj, 'n_eval', None))
        super(Solution, self).__setattr__('index', getattr(obj, 'index', None))
        self.var_name = getattr(obj, 'var_name', None)
        self.fitness_name = getattr(obj, 'fitness_name', None)
        self.verbose = getattr(obj, 'verbose', None)
        self.dim = getattr(obj, 'dim', None)
        self.N = getattr(obj, 'N', None)
        self.n_obj = getattr(obj, 'n_obj', None)
    
    @classmethod
    def from_dict(cls, x, space=None):
        if isinstance(x, dict):
            var_name = list(x.keys())
            res = cls.__new__(cls, x=list(x.values()), var_name=var_name)
        elif isinstance(x, list):
            var_name = list(x[0].keys())
            _x = [list(_.values()) for _ in x]
            res = cls.__new__(cls, x=_x, var_name=var_name)
        return res

    def to_dict(self, orient='index', with_index=False, space=None):
        # NOTE: avoid calling self.__getitem__
        # TODO: the following code only work 2D array
        obj = np.atleast_2d(self.view(np.ndarray))
        if orient == 'index':
            if with_index:
                res = {
                    _index : {
                        self.var_name[k] : obj[i, k] for k in range(self.dim)
                    } for i, _index in enumerate(self.index)
                } 
            else:
                res = [
                    {
                        self.var_name[k] : obj[i, k] for k in range(self.dim)
                    } for i, _index in enumerate(self.index)
                ]
                # if len(res) == 1:
                #     res = res[0]
        elif orient == 'var':
            if with_index:
                res = {
                    _name : {
                        i : obj[i, k] for i in self.index
                    } for k, _name in enumerate(self.var_name)
                }
            else:
                res = {
                    _name : list(obj[:, k]) for k, _name in enumerate(self.var_name)
                } 
        return res

    def __str__(self):
        var_name = self.var_name.tolist()
        headers = var_name + ['n_eval'] + self.fitness_name if self.verbose else var_name
        if len(self.shape) == 1:
            t = [self.tolist() + self.n_eval.tolist() + self.fitness.tolist()] \
                if self.verbose else [self.tolist()]
        else:
            t = np.c_[self, self.n_eval, self.fitness].tolist() \
                if self.verbose else self.tolist()
        
        return tabulate(t, headers=headers, showindex=self.index.tolist(), tablefmt='grid')

    def __repr__(self):
        return self.__str__()
                        
    def to_csv(self, fname, delimiter=',', append=False, header=True, index=True, 
               show_attr=True):
        var_name = self.var_name.tolist()
        if header:
            _header = var_name
            if index:
                _header = [''] + _header
            if show_attr:
                attr_name = ['n_eval'] + self.fitness_name
                _header += attr_name
            _header = ','.join(_header) + '\n'
        
        data = self.reshape(1, -1) if len(self.shape) == 1 else self
        if index:
            data = np.c_[self.index, data]
        if show_attr:
            data = np.c_[data, self.n_eval, self.fitness] 

        out = [','.join(map(str, row)) + '\n' for row in data.tolist()]
        mode = 'a' if append else 'w'
        with open(fname, mode) as f:
            if header:
                f.writelines(_header)
            f.writelines(out)


from setuptools import setup, find_packages

with open('README.md', 'r') as fh:
    long_description = fh.read()

setup(
    name="bayes-optim",
    version="0.1.0",
    author="Hao Wang",
    author_email="wangronin@gmail.com",
    packages=find_packages(),
    description="A Bayesian Optimization Library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/wangronin/Bayesian-Optimization",
    package_dir={'bayes_optim': 'bayes_optim'},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)"
    ],
    download_url='https://github.com/wangronin/Bayesian-Optimization/archive/v0.1.0.tar.gz',
    python_requires='>=3.6',
    install_requires=[ 
        'dill>=0.3.2',
        'joblib>=0.16.0',
        'numpy>=1.19.0',
        'pyDOE==0.3.8',
        'scikit-learn>=0.23.0',
        'scipy>=1.5.0',
        'sklearn==0.0',
        'tabulate>=0.8.7',
        'threadpoolctl>=2.1.0'
    ]
)


"""
Created on Mon Apr 23 17:16:39 2018

@author: Hao Wang
@email: wangronin@gmail.com

"""
import numpy as np
import os, sys, dill, functools, logging, time

from abc import ABC, abstractmethod
from copy import copy, deepcopy
from typing import Callable, Any, Tuple, Optional
from joblib import Parallel, delayed

from sklearn.metrics import r2_score
from sklearn.cluster import KMeans

from . import AcquisitionFunction
from .Solution import Solution
from .SearchSpace import SearchSpace
from .utils import arg_to_int
from .misc import LoggerFormatter
from .acquisition_optim import argmax_restart

class baseOptimizer(ABC):
    def __init__(self, verbose):
        self.verbose = verbose
        self.xopt = None 
        self.fopt = None 
        self.stop_dict = {}

    @abstractmethod
    def ask(self, n_point=None):
        """Get suggestions from the optimizer.

        Parameters
        ----------
        n_point : int
            Desired number of parallel suggestions in the output

        Returns
        -------
        next_guess : list of dict
            List of `n_suggestions` suggestions to evaluate the objective
            function. Each suggestion is a dictionary where each key
            corresponds to a parameter being optimized.
        """
        return

    @abstractmethod
    def tell(self, X, y):
        """Feed an observation back.

        Parameters
        ----------
        X : list of dict-like
            Places where the objective function has already been evaluated.
            Each suggestion is a dictionary where each key corresponds to a
            parameter being optimized.
        y : array-like, shape (n,)
            Corresponding values where objective has been evaluated
        """
        return

    @abstractmethod
    def evaluate(self, X):
        return 

    @abstractmethod
    def check_stop(self):
        return

    def step(self):
        X = self.ask()    
        func_vals = self.evaluate(X)
        self.tell(X, func_vals)
    
    def run(self):
        while not self.check_stop():
            self.step()
        return self.xopt, self.fopt, self.stop_dict

    @property
    def logger(self):
        return self._logger

    @logger.setter
    def logger(self, logger):
        self._logger = logging.getLogger(self.__class__.__name__)
        self._logger.setLevel(logging.DEBUG)
        fmt = LoggerFormatter()

        if self.verbose:
            # create console handler and set level to warning
            ch = logging.StreamHandler(sys.stdout)
            ch.setLevel(logging.INFO)
            ch.setFormatter(fmt)
            self._logger.addHandler(ch)

        # create file handler and set level to debug
        if logger is not None:
            fh = logging.FileHandler(logger)
            fh.setLevel(logging.DEBUG)
            fh.setFormatter(fmt)
            self._logger.addHandler(fh)

        if hasattr(self, 'logger'):
            self._logger.propagate = False


class baseBO(ABC):
    def __init__(
        self, 
        search_space: SearchSpace, 
        obj_fun: Callable,
        parallel_obj_fun: Callable = None,
        eq_fun: Callable = None, 
        ineq_fun: Callable = None, 
        model: Optional[Any] = None,   # TODO: regulate the type for `model`
        eval_type: str = 'list',
        DoE_size: Optional[int] = None, 
        warm_data: Tuple = (), 
        n_point: int = 1,
        acquisition_fun: str = 'EI',
        acquisition_par: dict = {},
        acquisition_optimization: dict = {},
        ftarget: Optional[float] = None,
        max_FEs: Optional[int] = None, 
        minimize: bool = True, 
        n_job: int = 1,
        data_file: Optional[str] = None, 
        verbose: bool = False, 
        random_seed: Optional[int] = None, 
        logger: Optional[str] = None,
        ):
        """ The base class for Bayesian Optimization

        Parameters
        ----------
        search_space : SearchSpace
            The search space, an instance of `SearchSpace` class.
        obj_fun : Callable
            The objective function to optimize.
        parallel_obj_fun : Callable, optional
            The objective function that takes multiple solutions simultaneously and 
            implement the parallelization by itself, by default None.
        eq_fun : Callable, optional
            The equality constraints, whose return value should have the same size as the 
            number of equality constraints, by default None.
        ineq_fun : Callable, optional
            The inequality constraints, whose return value should have the same size as
            the number of inequality constraints, by default None.
        model : Any, optional
            The surrogate mode, which will be automatically created if not passed in, 
            by default None.
        eval_type : str, optional
            The type of input argument allowed by `obj_func` or `parallel_obj_fun`: 
            it could be either 'list' or 'dict', by default 'list'.
        DoE_size : int, optional
            The size of inital Design of Experiment (DoE), by default None.
        warm_data: Tuple, optional
            The warm-starting data in a pair of design points and its objective values 
            `(X, y)`, where `X` should be a list of points and has the same length with `y`.
            When provided, the initial sampling (DoE) operation is skipped. 
        n_point : int, optional
            The number of candidate solutions proposed using infill-criteria, by default 1
        acquisition_fun : str, optional
            The acquisition function, by default 'EI'
        acquisition_par : dict, optional
            Extra parameters to the acquisition function, by default {}
        acquisition_optimization : dict, optional
            Additional parameters controlling the acquisition optimization, by default {}
        ftarget : float, optional
            The target value to hit, by default None
        max_FEs : int, optional
            The maximal number of evaluations, by default None
        minimize : bool, optional
            To minimize or maximize, by default True
        n_job : int, optional
            The number of allowable jobs for parallelizing the function evaluation 
            (if `parallel_obj_fun` is not specified) and Only Effective when n_point > 1, 
            by default 1
        data_file : str, optional
            The name of the file to store extra historical information during the run,
            by default None
        verbose : bool, optional
            The verbosity, by default False
        random_seed : int, optional
            The seed for pseudo-random number generators, by default None
        logger : str, optional
            Name of the logger file, by default None, which turns off the logging behaviour
        """        
        self.obj_fun = obj_fun
        self.parallel_obj_fun = parallel_obj_fun
        self.h = eq_fun
        self.g = ineq_fun
        self.n_job = max(1, int(n_job))
        self.n_point = max(1, int(n_point))
        self.ftarget = ftarget 
        self.minimize = minimize
        self.verbose = verbose
        self.data_file = data_file
        self.max_FEs = int(max_FEs) if max_FEs else np.inf  

        self.search_space = search_space
        self.DoE_size = DoE_size

        self.acquisition_fun = acquisition_fun
        self._acquisition_par = acquisition_par
        self._acquisition_callbacks = []   # the callback functions executed after 
                                           # every call of `arg_max_acquisition`
        self.model = model
        self.logger = logger
        self.random_seed = random_seed

        self._set_internal_optimization(**acquisition_optimization)
        self._get_best = np.min if self.minimize else np.max
        self._eval_type = eval_type   
        self._init_flatfitness_trial = 2
        self._set_aux_vars()
        self.warm_data = warm_data
    
    @property
    def acquisition_fun(self):
        return self._acquisition_fun

    @acquisition_fun.setter
    def acquisition_fun(self, fun):
        if isinstance(fun, str):
            self._acquisition_fun = fun
        else:
            assert hasattr(fun, '__call__')
        self._acquisition_fun = fun

    @property
    def DoE_size(self):
        return self._DoE_size

    @DoE_size.setter
    def DoE_size(self, DoE_size):
        if DoE_size:
            if isinstance(DoE_size, str):
                self._DoE_size = int(eval(DoE_size))
            elif isinstance(DoE_size, (int, float)):
                self._DoE_size = int(DoE_size)
            else: 
                raise ValueError
        else:
            self._DoE_size = int(self.dim * 5)

    @property
    def random_seed(self):
        return self._random_seed
    
    @random_seed.setter
    def random_seed(self, seed):
        if seed:
            self._random_seed = int(seed)
            if self._random_seed:
                np.random.seed(self._random_seed)

    @property
    def search_space(self):
        return self._search_space

    @search_space.setter
    def search_space(self, search_space):
        self._search_space = search_space
        self.dim = len(self._search_space)
        self.var_names = self._search_space.var_name
        self.r_index = self._search_space.id_C       # indices of continuous variable
        self.i_index = self._search_space.id_O       # indices of integer variable
        self.d_index = self._search_space.id_N       # indices of categorical variable

        self.param_type = self._search_space.var_type
        self.N_r = len(self.r_index)
        self.N_i = len(self.i_index)
        self.N_d = len(self.d_index)

    @property
    def warm_data(self):
        return self._warm_data
    
    @warm_data.setter
    def warm_data(self, data):
        assert self.iter_count == 0  # warm data should only be provided in the begining
        if data is None or len(data) == 0:
            self._warm_data = None
        else:
            X, y = data
            assert len(X) == len(y)

            if isinstance(X, Solution):
                assert X.var_name == self.var_names
            else:
                X = self._to_geno(X)

            X.fitness = y
            X.n_eval = 1
            assert all([isinstance(_, float) for _ in X[:, self.r_index].ravel()])
            assert all([isinstance(_, int) for _ in X[:, self.i_index].ravel()])
            assert all([isinstance(_, str) for _ in X[:, self.d_index].ravel()])
            self._warm_data = X
            self.tell(X, y, warm_start=True)

    @property
    def logger(self):
        return self._logger

    @logger.setter
    def logger(self, logger):
        if isinstance(logger, logging.Logger):
            self._logger = logger
            self._logger.propagate = False
            return

        self._logger = logging.getLogger(self.__class__.__name__)
        self._logger.setLevel(logging.DEBUG)
        fmt = LoggerFormatter()

        if self.verbose:
            # create console handler and set level to warning
            ch = logging.StreamHandler(sys.stdout)
            ch.setLevel(logging.INFO)
            ch.setFormatter(fmt)
            self._logger.addHandler(ch)

        # create file handler and set level to debug
        if logger is not None:
            fh = logging.FileHandler(logger)
            fh.setLevel(logging.DEBUG)
            fh.setFormatter(fmt)
            self._logger.addHandler(fh)

        if hasattr(self, 'logger'):
            self._logger.propagate = False
    
    def _set_aux_vars(self):
        self.iter_count = 0
        self.eval_count = 0
        self.stop_dict = {}
        self.hist_f = []

        if self._eval_type == 'list':
            self._to_pheno = lambda x: x.tolist()
            self._to_geno = lambda x: Solution(x, var_name=self.var_names)
        elif self._eval_type == 'dict':
            self._to_pheno = lambda x: x.to_dict(space=self._search_space)
            self._to_geno = lambda x: Solution.from_dict(x, space=self._search_space)

    def _set_internal_optimization(self, **kwargs):
        if 'optimizer' in kwargs:
            self._optimizer = kwargs['optimizer']
        else:
            if self.N_d + self.N_i == 0 and hasattr(self.model, 'gradient'):
                self._optimizer = 'BFGS'
            else:
                self._optimizer = 'MIES'

        # NOTE: `AQ`: acquisition
        if 'max_FEs' in kwargs:
            self.AQ_max_FEs = arg_to_int(kwargs['max_FEs'])
        else:
            self.AQ_max_FEs = int(5e2 * self.dim) if self._optimizer == 'MIES' else \
                int(1e2 * self.dim)
        
        self.AQ_n_restart = int(5 * self.dim) if 'n_restart' not in kwargs \
            else arg_to_int(kwargs['n_restart'])
        self.AQ_wait_iter = 3 if 'wait_iter' not in kwargs \
            else arg_to_int(kwargs['wait_iter'])

        self._argmax_restart = functools.partial(
            argmax_restart, 
            search_space=self._search_space,
            h=self.h,
            g=self.g,
            eval_budget=self.AQ_max_FEs,
            n_restart=self.AQ_n_restart,
            wait_iter=self.AQ_wait_iter,
            optimizer=self._optimizer
        ) 

    def _check_params(self):
        # TODO: add more parameter check-ups
        if np.isinf(self.max_FEs):
            raise ValueError('max_FEs cannot be infinite')

        assert hasattr(AcquisitionFunction, self._acquisition_fun)
    
    def _compare(self, f1, f2):
        """Test if objecctive value f1 is better than f2
        """
        return f1 < f2 if self.minimize else f2 > f1

    def run(self):
        while not self.check_stop():
            self.step()
        return self.xopt, self.fopt, self.stop_dict

    def step(self):
        X = self.ask()

        t0 = time.time()
        func_vals = self.evaluate(X)
        self._logger.info('evaluation takes {:.4f}s'.format(time.time() - t0))

        self.tell(X, func_vals)

    def ask(self, n_point=None):
        if self.model.is_fitted:
            n_point = self.n_point if n_point is None else self.n_point
            X = self.arg_max_acquisition(n_point=n_point)
            X = self._search_space.round(X)  # round to precision if specified

            X = Solution(
                X, index=len(self.data) + np.arange(len(X)), 
                var_name=self.var_names
            )
            X = self.pre_eval_check(X)

            # TODO: handle the constraints when performing the random sampling
            if len(X) < n_point:
                self._logger.warn(
                    "iteration {}: duplicated solution found " 
                    "by optimization! New points is taken from random "
                    "design".format(self.iter_count)
                )
                N = n_point - len(X)
                method = 'LHS' if N > 1 else 'uniform'
                s = self._search_space.sampling(N=N, method=method) 
                X = X.tolist() + s
                X = Solution(
                    X, index=len(self.data) + np.arange(len(X)), 
                    var_name=self.var_names
                )
                X = self._search_space.round(X)
        else:   # initial DoE
            if not n_point:
                n_point = self._DoE_size
                
            X = self._search_space.round(
                self.create_DoE(n_point)
            )
        return self._to_pheno(X)
    
    def tell(self, X, func_vals, warm_start=False):
        """Tell the BO about the function values of proposed candidate solutions

        Parameters
        ----------
        X : List of Lists or Solution
            The candidate solutions which are usually proposed by the `self.ask` function
        func_vals : List/np.ndarray of reals
            The corresponding function values
        """
        X = self._to_geno(X)

        if warm_start:
            msg = 'warm-starting from {} points:'.format(len(X))
        elif self.iter_count == 0:
            msg = 'initial DoE of size {}:'.format(len(X))
        else:
            msg = 'iteration {}, {} infill points:'.format(self.iter_count, len(X))

        self._logger.info(msg)
        X_ = self._to_pheno(X)
        
        for i in range(len(X)):
            X[i].fitness = func_vals[i]
            X[i].n_eval += 1
            
            if not warm_start:
                self.eval_count += 1

            self._logger.info(
                '#{} - fitness: {}, solution: {}'.format(
                    i + 1, func_vals[i], X_[i]
                )
            )

        X = self.post_eval_check(X)
        self.data = self.data + X if hasattr(self, 'data') else X

        if self.data_file is not None:
            X.to_csv(self.data_file, header=False, append=True)

        self.fopt = self._get_best(self.data.fitness)
        _xopt = self.data[np.where(self.data.fitness == self.fopt)[0][0]] 
        self.xopt = self._to_pheno(_xopt)
        if self._eval_type == 'dict':
            self.xopt = self.xopt[0]

        self._logger.info('fopt: {}'.format(self.fopt))   
        self._logger.info('xopt: {}'.format(self.xopt)) 

        if not self.model.is_fitted: 
            self._fBest_DoE = copy(self.fopt) # the best point in the DoE 
            self._xBest_DoE = copy(self.xopt)

        r2 = self.update_model()   
        self._logger.info('Surrogate model r2: {}\n'.format(r2))

        if not warm_start:
            self.iter_count += 1
            self.hist_f.append(self.fopt)

    def create_DoE(self, n_point=None):
        DoE = []
        while len(DoE) < n_point:
            DoE += self._search_space.sampling(n_point - len(DoE), method='LHS')
            DoE = self.pre_eval_check(DoE).tolist()
        
        return Solution(DoE, var_name=self.var_names)

    @abstractmethod
    def pre_eval_check(self, X):
        raise NotImplementedError
    
    def post_eval_check(self, X):
        _ = np.isnan(X.fitness) | np.isinf(X.fitness)
        if np.any(_):
            self._logger.warn(
                '{} candidate solutions are removed '
                'due to falied fitness evaluation: \n{}'.format(sum(_), str(X[_, :]))
            )
            X = X[~_, :] 
        return X
        
    def evaluate(self, X):
        """Evaluate the candidate points and update evaluation info in the dataframe
        """
        # Parallelization is handled by the objective function itself
        if self.parallel_obj_fun is not None:  
            func_vals = self.parallel_obj_fun(X)
        else: 
            if self.n_job > 1: # or by ourselves..
                func_vals = Parallel(n_jobs=self.n_job)(delayed(self.obj_fun)(x) for x in X)
            else:              # or sequential execution
                func_vals = [self.obj_fun(x) for x in X]
                
        return func_vals

    def update_model(self):
        # TODO: implement a proper model selection here 
        # TODO: in case r2 is really poor, re-fit the model or log-transform `fitness`?
        data = self.data
        fitness = data.fitness

        # to standardize the response values to prevent numerical overflow that might 
        # appear in the MGF-based acquisition function.
        # Standardization should make it easier to specify the GP prior, compared to 
        # rescaling values to the unit interval.
        fitness_ = (fitness - np.mean(fitness)) / np.std(fitness) \
            if self._acquisition_fun == 'MGFI' else fitness

        self.fmin, self.fmax = np.min(fitness_), np.max(fitness_)
        self.frange = self.fmax - self.fmin

        self.model.fit(data, fitness_)
        fitness_hat = self.model.predict(data)

        # TODO: need to report more performance metric for regression   
        r2 = r2_score(fitness_, fitness_hat)
        return r2

    def arg_max_acquisition(self, n_point=None, return_value=False):
        """
        Global Optimization of the acqusition function / Infill criterion
        Returns
        -------
            candidates: tuple of list,
                candidate solution (in list)
            values: tuple,
                criterion value of the candidate solution
        """
        self._logger.debug('acquisition optimziation...')
        t0 = time.time()
        n_point = self.n_point if n_point is None else int(n_point)
        return_dx = True if self._optimizer == 'BFGS' else False

        if n_point > 1:  # multi-point/batch sequential strategy
            candidates, values = self._batch_arg_max_acquisition(
                n_point=n_point, return_dx=return_dx
            )
        else:            # single-point strategy
            criteria = self._create_acquisition(par={}, return_dx=return_dx)
            candidates, values = self._argmax_restart(criteria, logger=self._logger)
            candidates, values = [candidates], [values]

        self._logger.debug(
            'acquisition optimziation takes {:.4f}s'.format(time.time() - t0)
        )
        for callback in self._acquisition_callbacks: callback()

        return (candidates, values) if return_value else candidates

    def _create_acquisition(self, fun=None, par={}, return_dx=False):
        fun = fun if fun is not None else self._acquisition_fun
        par = copy(self._acquisition_par) if not par else par
        par.update({'model' : self.model, 'minimize' : self.minimize})

        criterion = getattr(AcquisitionFunction, fun)(**par)
        return functools.partial(criterion, return_dx=return_dx)

    def _batch_arg_max_acquisition(self, n_point, return_dx):
        raise NotImplementedError

    def check_stop(self):
        if self.eval_count >= self.max_FEs:
            self.stop_dict['max_FEs'] = self.eval_count
        
        if self.ftarget is not None and hasattr(self, 'xopt'):
            if self._compare(self.fopt, self.ftarget):
                self.stop_dict['ftarget'] = self.fopt

        return bool(self.stop_dict)
    
    def save(self, filename):
        with open(filename, 'wb') as f:
            # NOTE: we need to dump `self.data` first. Otherwise, some 
            # attributes of it will be lost
            if hasattr(self, 'data'):
                self.data = dill.dumps(self.data)
            
            if len(self._logger.handlers) > 1:
                _ = [h for h in self._logger.handlers if isinstance(h, logging.FileHandler)]
                _logger = _[0].baseFilename
            else: 
                _logger = None
                        
            logger = self._logger
            self._logger = _logger

            dill.dump(self, f)

            self._logger = logger
            if hasattr(self, 'data'):
                self.data = dill.loads(self.data)
        
    @classmethod
    def load(cls, filename):
        with open(filename, 'rb') as f:
            obj = dill.load(f)
            if hasattr(obj, 'data'):
                obj.data = dill.loads(obj.data)

            obj.logger = obj._logger
        return obj
    



import numpy as np
import logging, os, random, string, re
from copy import copy

# TODO: re-written those functions to C/Cython
def non_dominated_set_2d(y, minimize=True):
    """ 
    Argument
    --------
    y : numpy 2d array, 
        where the each solution occupies a row
    """
    y = np.asarray(y)
    N, _ = y.shape

    if isinstance(minimize, bool):
        minimize = [minimize]
    
    minimize = np.asarray(minimize).ravel()
    assert len(minimize) == 1 or minimize.shape == (N, ) 
    y *= (np.asarray([-1] * N) ** minimize).reshape(-1, 1)

    _ = np.argsort(y[:, 0])[::-1]
    y2 = y[_, 1]
    ND = []
    for i in range(N):
        v = y2[i]
        if not any(v <= y2[ND]) or len(ND) == 0:
            ND.append(i) 
    return _[ND]

def non_dominated_set_3d(y, minimize=True):
    pass

def fast_non_dominated_sort(fitness):
    fronts = []
    dominated_set = []
    mu = fitness.shape[1]
    n_domination = np.zeros(mu)
    
    for i in range(mu):
        p = fitness[:, i]
        p_dominated_set = []
        n_p = 0
        
        for j in range(mu):
            q = fitness[:, j]
            if i != j:
                # TODO: verify this part 
                # check the strict domination
                # allow for duplication points on the same front
                if all(p <= q) and not all(p == q):
                    p_dominated_set.append(j)
                elif all(p >= q) and not all(p == q):
                    n_p += 1
                    
        dominated_set.append(p_dominated_set)
        n_domination[i] = n_p
    
    # create the first front
    fronts.append(np.nonzero(n_domination == 0)[0].tolist())
    n_domination[n_domination == 0] = -1
    
    i = 0
    while True:
        for p in fronts[i]:
            p_dominated_set = dominated_set[p]
            n_domination[p_dominated_set] -= 1
            
        _front = np.nonzero(n_domination == 0)[0].tolist()
        n_domination[n_domination == 0] = -1
            
        if len(_front) == 0:
            break
        fronts.append(_front)
        i += 1
        
    return fronts

# TODO: implement this as a C procedure
def proportional_selection(perf, N, minimize=True, replacement=True):
    def select(perf):
        perf_min = np.min(perf)
        interval = np.cumsum((perf - perf_min) / (np.sum(perf) - perf_min * len(perf)))
        return np.nonzero(np.random.rand() <= interval)[0][0]
    
    perf = np.array(perf)
    if minimize:
        perf = -perf
        perf -= np.min(perf)

    if replacement:
        res = [select(perf) for i in range(N)]
    else:
        assert N <= len(perf)
        perf_ = copy(perf)
        idx = list(range(0, len(perf)))
        res = []
        for i in range(N):
            if len(perf_) == 1:
                res.append(idx[0])
            else:
                _ = select(perf_)
                res.append(idx[_])
                perf_ = np.delete(perf_, _)
                del idx[_]
    return res

# TODO: double check this one. It causes the explosion of step-sizes in MIES
def handle_box_constraint(x, lb, ub):
    """This function transforms x to t w.r.t. the low and high
    boundaries lb and ub. It implements the function T^{r}_{[a,b]} as
    described in Rui Li's PhD thesis "Mixed-Integer Evolution Strategies
    for Parameter Optimization and Their Applications to Medical Image 
    Analysis" as alorithm 6.
    
    """    
    x = np.asarray(x, dtype='float')
    shape_ori = x.shape
    x = np.atleast_2d(x)
    lb = np.atleast_1d(lb)
    ub = np.atleast_1d(ub)
    
    transpose = False
    if x.shape[0] != len(lb):
        x = x.T
        transpose = True
    
    lb, ub = lb.flatten(), ub.flatten()
    lb_index = np.isfinite(lb)
    up_index = np.isfinite(ub)
    
    valid = np.bitwise_and(lb_index, up_index)
    
    LB = lb[valid][:, np.newaxis]
    UB = ub[valid][:, np.newaxis]

    y = (x[valid, :] - LB) / (UB - LB)
    I = np.mod(np.floor(y), 2) == 0
    yprime = np.zeros(y.shape)
    yprime[I] = np.abs(y[I] - np.floor(y[I]))
    yprime[~I] = 1.0 - np.abs(y[~I] - np.floor(y[~I]))

    x[valid, :] = LB + (UB - LB) * yprime
    
    if transpose:
        x = x.T
    return x.reshape(shape_ori)

class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

# TODO: use relative path for %(pathname)s
class LoggerFormatter(logging.Formatter):
    default_time_format = '%m/%d/%Y %H:%M:%S'
    default_msec_format = '%s,%02d'

    FORMATS = {
        logging.DEBUG : '%(asctime)s - [%(name)s.%(levelname)s] {%(pathname)s:%(lineno)d} -- %(message)s',
        logging.INFO : '%(asctime)s - [%(name)s.%(levelname)s] -- %(message)s',
        logging.WARNING : '%(asctime)s - [%(name)s.%(levelname)s] {%(name)s} -- %(message)s',
        logging.ERROR : '%(asctime)s - [%(name)s.%(levelname)s] {%(name)s} -- %(message)s',
        'DEFAULT' : '%(asctime)s - %(levelname)s -- %(message)s'
    }
    
    def __init__(self, fmt='%(asctime)s - %(levelname)s -- %(message)s'):
        LoggerFormatter.FORMATS['DEFAULT'] = fmt
        super().__init__(fmt=fmt, datefmt=None, style='%') 
    
    def format(self, record):

        # Save the original format configured by the user
        # when the logger formatter was instantiated
        _fmt = self._style._fmt

        # Replace the original format with one customized by logging level
        self._style._fmt = self.FORMATS.get(record.levelno, self.FORMATS['DEFAULT'])

        # Call the original formatter class to do the grunt work
        fmt = logging.Formatter.format(self, record)

        # Restore the original format configured by the user
        self._style._fmt = _fmt
        return fmt

def random_string(k=15):
    return ''.join(random.choices(string.ascii_letters + string.digits, k=15))

def expand_replace(s):
    m = re.match(r'${.*}', s)
    for _ in m.group():
        s.replace(_, os.path.expandvars(_))
    return s

if __name__ == '__main__':
    # TODO: this goes to unittest
    np.random.seed(1)
    perf = np.random.randn(20)
    print(perf)
    print(proportional_selection(perf, 20, minimize=False, replacement=False))
